=head1 NAME

Text::Gr2::FeatureVal - values for a feature

=head1 DESCRIPTION

No user serviceable parts inside. Use via L<Text::Gr2::FeatureRef>
instead.

=head1 METHODS

=head2 clone

Produce another one, just like this.

=cut

1;
