using System;
using NUnit.Framework;
using NGraphite;
using System.Text;
using System.Runtime.InteropServices;

namespace NGraphiteTests
{
	public class TestConstants
	{
		public const string PaduakFontLocation = "/usr/share/fonts/truetype/padauk/Padauk.ttf";
		public const string DejaVuSansLocation = "/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf";
	}
}
